class FlaskPasetoError(Exception):
    """
    Base class for all exceptions.
    """

    pass


class ConfigError(FlaskPasetoError):
    """
    An Exception occurred when the configurations are not completed.
    """

    pass


class EncodeError(FlaskPasetoError):
    """
    An Exception occurred when a PASETO encoding process failed.
    """

    pass


class DecodeError(FlaskPasetoError):
    """
    An Exception occurred when a PASETO decoding process failed.
    """

    pass



from .cookie_session import PasetoCookieSessionInterface
from .exceptions import DecodeError, EncodeError, FlaskPasetoError
from .issuer import PasetoIssuer
from .login_manager import PasetoLoginManager
from .token import Token, current_paseto, paseto_required
from .verifier import PasetoVerifier

__version__ = "0.4.0"
__title__ = "Flask PASETO Extended"
__description__ = "PASETO (Platform-Agnostic Security Tokens) for Flask applications."
__url__ = "https://github.com/dajiaji/flask-paseto-extended"
__uri__ = __url__
__doc__ = __description__ + " <" + __uri__ + ">"
__author__ = "Ajitomi Daisuke"
__email__ = "dajiaji@gmail.com"
__license__ = "MIT"
__copyright__ = "Copyright 2021 Ajitomi Daisuke"
__all__ = [
    "PasetoCookieSessionInterface",
    "PasetoLoginManager",
    "PasetoIssuer",
    "PasetoVerifier",
    "Token",
    "current_paseto",
    "paseto_required",
    "FlaskPasetoError",
    "EncodeError",
    "DecodeError",
]



import json
import typing as t

from pyseto import Key, Paseto

from .config import PASETO_VERSIONS_ACCEPTABLE
from .exceptions import EncodeError

PASETO_DEFAULT_USE_ISS: bool = True
PASETO_DEFAULT_USE_IAT: bool = False
PASETO_DEFAULT_EXP: int = 3600
PASETO_DEFAULT_USE_KID: bool = False
PASETO_DEFAULT_SERIALIZER: t.Any = json


class PasetoIssuer(object):
    def __init__(self, app=None, add_context_processor=True):

        if app is not None:
            self.init_app(app, add_context_processor)

    def init_app(self, app, add_context_processor=True):
        """
        Configures an Flask application to use this PasetoIssuer.
        """
        app.paseto_issuer = self

        # _iss
        self._iss = app.config.get("PASETO_ISS", "")
        if not self._iss:
            raise ValueError("PASETO_ISS must be set.")
        if not isinstance(self._iss, str):
            raise ValueError("PASETO_ISS must be str.")

        # _use_iss
        self._use_iss = app.config.get("PASETO_USE_ISS", PASETO_DEFAULT_USE_ISS)
        if not isinstance(self._use_iss, bool):
            raise ValueError("PASETO_USE_ISS must be bool.")

        # _use_iat
        self._use_iat = app.config.get("PASETO_USE_IAT", PASETO_DEFAULT_USE_IAT)
        if not isinstance(self._use_iat, bool):
            raise ValueError("PASETO_USE_IAT must be bool.")

        # _exp
        self._exp = app.config.get("PASETO_EXP", PASETO_DEFAULT_EXP)
        if not isinstance(self._exp, int) or self._exp < 0:
            raise ValueError("PASETO_EXP must be int (>= 0).")

        # _use_kid
        self._use_kid = app.config.get("PASETO_USE_KID", PASETO_DEFAULT_USE_KID)
        if not isinstance(self._use_kid, bool):
            raise ValueError("PASETO_USE_KID must be bool.")

        # _serializer
        self._serializer: t.Any = app.config.get("PASETO_SERIALIZER", PASETO_DEFAULT_SERIALIZER)
        if not hasattr(self._serializer, "dumps") or not callable(self._serializer.dumps):
            raise ValueError("PASETO_SERIALIZER must have a callable 'dumps'.")

        # _keys
        keys: list = app.config.get("PASETO_PRIVATE_KEYS", [])
        if not keys:
            raise ValueError("PASETO_PRIVATE_KEYS must be set.")
        self._keys: dict = {}
        key: t.Any = None
        kid: bytes = b""
        for k in keys:
            if "paserk" in k:
                try:
                    key = Key.from_paserk(k["paserk"])
                except Exception as err:
                    raise ValueError("Invalid PASERK data.") from err
                if key.purpose == "local":
                    raise ValueError("A local key is not allowed.")
            else:
                if "version" not in k:
                    raise ValueError("A key object must have a 'paserk' or a pair of 'version' and 'key'.")
                if not isinstance(k["version"], int):
                    raise ValueError("A 'version' in PASETO_PRIVATE_KEYS must be int.")
                if k["version"] not in PASETO_VERSIONS_ACCEPTABLE:
                    raise ValueError(f"Invalid PASETO version: {k['version']}.")
                if "key" not in k:
                    raise ValueError("A key object must have a 'paserk' or a pair of 'version' and 'key'.")
                try:
                    key = Key.new(k["version"], "public", k["key"])
                except Exception as err:
                    raise ValueError("A 'key' must be a PEM formatted key.") from err
            kid = key.to_paserk_id()
            self._keys[kid] = {"key": key}

        self._paseto = Paseto.new(exp=self._exp, include_iat=self._use_iat)
        return

    def issue(self, payload: dict, kid: str = "") -> str:

        key = self._keys[list(self._keys)[0]] if len(self._keys) == 1 else self._keys.get(kid, None)
        if not key:
            raise ValueError("A signing key is not found.")

        footer: dict = {}
        if self._use_iss:
            payload["iss"] = self._iss
        if self._use_kid:
            footer["kid"] = list(self._keys.keys())[0] if len(self._keys) == 1 else kid
        try:
            if not footer:
                return self._paseto.encode(key["key"], payload, serializer=self._serializer)
            return self._paseto.encode(key["key"], payload, footer, serializer=self._serializer)
        except Exception as err:
            raise EncodeError("Failed to encode a token.") from err



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['flask_paseto_extended']

package_data = \
{'': ['*']}

install_requires = \
['Flask-Login>=0.5.0,<0.6.0', 'Flask>=2.0.2,<3.0.0', 'pyseto>=1.5.0,<2.0.0']

extras_require = \
{':extra == "docs"': ['Sphinx[docs]>=4.3.1,<5.0.0',
                      'sphinx-rtd-theme[docs]>=1.0.0,<2.0.0',
                      'sphinx-autodoc-typehints[docs]>=1.12.0,<2.0.0']}

setup_kwargs = {
    'name': 'flask-paseto-extended',
    'version': '0.4.0',
    'description': 'PASETO (Platform-Agnostic Security Tokens) for Flask applications.',
    'long_description': '# Flask PASETO Extended\n\n![Flask PASETO Extended](https://github.com/dajiaji/flask-paseto-extended/wiki/images/flask_paseto_extended_header.png)\n\n[![PyPI version](https://badge.fury.io/py/flask-paseto-extended.svg)](https://badge.fury.io/py/flask-paseto-extended)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/flask-paseto-extended)\n[![Documentation Status](https://readthedocs.org/projects/flask-paseto-extended/badge/?version=latest)](https://flask-paseto-extended.readthedocs.io/en/latest/?badge=latest)\n![Github CI](https://github.com/dajiaji/flask-paseto-extended/actions/workflows/python-package.yml/badge.svg)\n[![codecov](https://codecov.io/gh/dajiaji/flask-paseto-extended/branch/main/graph/badge.svg?token=QN8GXEYEP3)](https://codecov.io/gh/dajiaji/flask-paseto-extended)\n\nFlask-PASETO-Extended provides following four classes to use [PASETO (Platform-Agnostic Security Tokens)](https://paseto.io/) for Flask applications:\n\n- **PasetoIssuer**\n  - This class can be used for issuing `public` (signed) PASETO. It is suitable for using PASETO as API tokens.\n- **PasetoVerifier**\n  - This class can be used for verifying `public` (signed) PASETO. It is suitable for using PASETO as API tokens.\n- **PasetoCookieSessionInterface**\n  - Flask (`Flask.sessions`) stores session information as a Cookie value. By using this class, you can serialize the session information as a `local` (encrypted and then MACed) PASETO.\n- **PasetoLoginManager**\n  - By using this class together with [Flask-Login](https://github.com/maxcountryman/flask-login), you can use a `local` PASETO for remember-me tokens which is also encoded into a Cookie value.\n\nFor encoding/decoding PASETO, we have adopted [PySETO](https://github.com/dajiaji/pyseto),\nwhich is a PASETO/PASERK implementation supporting all of PASETO versions (\n[v4](https://github.com/paseto-standard/paseto-spec/blob/master/docs/01-Protocol-Versions/Version4.md),\n[v3](https://github.com/paseto-standard/paseto-spec/blob/master/docs/01-Protocol-Versions/Version3.md),\n[v2](https://github.com/paseto-standard/paseto-spec/blob/master/docs/01-Protocol-Versions/Version2.md) and\n[v1](https://github.com/paseto-standard/paseto-spec/blob/master/docs/01-Protocol-Versions/Version1.md)) and purposes (`local` and `public`).\n\n## Index\n- [Installation](#installation)\n- [Basic Usage](#basic-usage)\n  - [PasetoIssuer/PasetoVerifier](#pasetoissuerverifier)\n  - [PasetoCookieSessionInterface](#pasetocookiesessioninterface)\n  - [PasetoLoginManager](#pasetologinmanager)\n- [API Reference](#api-reference)\n- [Tests](#tests)\n- [Contributing](#contributing)\n\n## Installation\n\nYou can install Flask-PASETO-Extended with pip:\n\n```py\n$ pip install flask-paseto-extended\n```\n\n## Basic Usage\n\nFlask-PASETO-Extended provides three classes for each purpose.\n\n### PasetoIssuer/Verifier\n\n`PasetoIssuer` and `PasetoVerifier` can be used for issuing and verifying `public` (signed) PASETO tokens.\n\nBy using `PasetoIssuer`, you can easily implement the endpoint issuing PASETO tokens as follows:\n\n```py\nimport flask\n\nfrom flask_paseto_extended import PasetoIssuer\n\n# Mock user database.\nusers = {"foo@bar.example": {"password": "mysecret"}}\n\n\napp = flask.Flask(__name__)\n\napp.config["PASETO_ISS"] = "https://issuer.example"\napp.config["PASETO_PRIVATE_KEYS"] = [\n    {\n        "version": 4,\n        "key": "-----BEGIN PRIVATE KEY-----\\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\\n-----END PRIVATE KEY-----",\n    },\n    # PASERK can also be used (RECOMMENDED).\n    # {\n    #     "paserk": "k4.secret.tMv7Q99M4hByfZU-SnEzB_oZu32fhQQUONnhG5QqN3Qeudu7vAR8A_1wYE4AcfCYfhayi3VyJcEfAEFdDiCxog",\n    # },\n]\nissuer = PasetoIssuer(app)\n\n\n@app.route("/login", methods=["POST"])\ndef login():\n    email = flask.request.form["email"]\n    if flask.request.form["password"] != users[email]["password"]:\n        return "Bad login"\n\n    token = issuer.issue(payload={"user": {"email": email}})\n    resp = flask.redirect(flask.url_for("protected_me"))\n    resp.set_cookie("paseto", token, httponly=True)  # Note: MUST add secure=True in production\n    return resp\n```\n\nOn the other hand, by using `PasetoVerifier`, you can easily implement the endpoint verifying PASETO tokens. You can enable PASETO token verification in your APIs by simply adding `@paseto_required` decorator to the API definitions. In the APIs, you can refer to the veified PASETO token with `current_paseto`.\n\n```py\nimport flask\nfrom flask import jsonify, make_response\n\nfrom flask_paseto_extended import PasetoVerifier, current_paseto, paseto_required\n\n# Mock user database.\nusers = {"foo@bar.example": {"password": "mysecret"}}\n\napp = flask.Flask(__name__)\n\n# Configurations for PasetoVerifier.\napp.config["PASETO_PUBLIC_KEYS"] = [\n    {\n        "iss": "https://issuer.exmaple",\n        "version": 4,\n        "key": "-----BEGIN PUBLIC KEY-----\\nMCowBQYDK2VwAyEAHrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI=\\n-----END PUBLIC KEY-----",\n    },\n    # PASERK can also be used (RECOMMENDED).\n    # {\n    #     "iss": "https://issuer.exmaple",\n    #     "paserk": "k4.public.Hrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI",\n    # },\n]\nverifier = PasetoVerifier(app)\n\n\n@verifier.token_loader\ndef token_loader(req: flask.Request):\n    # You must implement a callback func to extract a PASETO token from each request.\n    return req.cookies.get("paseto", None)\n\n\n@verifier.verification_error_handler\ndef verification_error_handler():\n    # You must also implement a callback func to handle token verification errors..\n    resp = make_response("Unauthorized")\n    resp.delete_cookie("paseto", httponly=True)\n    return resp\n\n\n@app.route("/protected/me")\n@paseto_required()\ndef protected_me():\n    return jsonify(current_paseto.payload["user"])\n```\n\nSee [examples/issuer_and_verifier.py](https://github.com/dajiaji/flask-paseto-extended/blob/main/examples/issuer_and_verifier.py) for a sample code that actually works.\n\n\n### PasetoCookieSessionInterface\n\nFlask (`Flask.sessions`) stores session information as a Cookie value. By using this class, you can serialize the session information as an encrypted (and then MACed) PASETO.\n\nThis class can be used as follows:\n\n```py\nimport flask\nfrom flask_paseto_extended import PasetoCookieSessionInterface\n\napp = flask.Flask(__name__)\napp.secret_key = "super secret string"\n\n# Use PASETO("v4" by default) for cookie sessions.\napp.session_interface = PasetoCookieSessionInterface()\n```\n\nSee [examples/cookie_session.py](https://github.com/dajiaji/flask-paseto-extended/blob/main/examples/cookie_session.py) for a sample code that actually works.\n\n### PasetoLoginManager\n\nBy using this class together with [Flask-Login](https://github.com/maxcountryman/flask-login), you can use PASETO for remember-me tokens which is also encoded into a Cookie value.\n\nThis class can be used as follows:\n\n```py\nimport flask\nimport flask_login\n# Import PasetoLoginManager instead of flask_login.LoginManager.\nfrom flask_paseto_extended import PasetoLoginManager\n\napp = flask.Flask(__name__)\napp.secret_key = "super secret string"\n\nlogin_manager = PasetoLoginManager(app)\n```\n\nSee [examples/login_manager.py](https://github.com/dajiaji/flask-paseto-extended/blob/main/examples/login_manager.py) for a sample code that actually works.\n\n## API Reference\n\nSee [Documentation](https://flask-paseto-extended.readthedocs.io/en/stable/api.html).\n\n\n## Tests\n\nYou can run tests from the project root after cloning with:\n\n```sh\n$ tox\n```\n\n## Contributing\n\nWe welcome all kind of contributions, filing issues, suggesting new features or sending PRs.\n',
    'author': 'Ajitomi Daisuke',
    'author_email': 'dajiaji@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/dajiaji/flask-paseto-extended',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



import typing as t
from functools import wraps

from flask import _request_ctx_stack, current_app, has_request_context
from werkzeug.local import LocalProxy

from .exceptions import ConfigError


class Token(object):
    """
    The parsed token object which can be referred as ``current_paseto`` in the
    views decorated by ``@token_required``.
    """

    def __init__(
        self,
        is_verified: bool,
        version: str = "",
        purpose: str = "",
        payload: t.Union[bytes, dict] = b"",
        footer: t.Union[bytes, dict] = b"",
        error: t.Optional[Exception] = None,
    ):

        self._is_verified = is_verified
        self._version = version
        self._purpose = purpose
        self._payload = payload
        self._footer = footer
        self._error = error
        self._header = (version + "." + purpose + ".").encode("utf-8")
        return

    @property
    def is_verified(self) -> bool:
        """
        If it is ``True``, the token has been verified.
        """
        return self._is_verified

    @property
    def version(self) -> str:
        """
        The version of the token. It will be ``"v1"``, ``"v2"``, ``"v3"`` or ``"v4"``.
        """
        return self._version

    @property
    def purpose(self) -> str:
        """
        The purpose of the token. It will be ``"local"`` or ``"public"``.
        """
        return self._purpose

    @property
    def header(self) -> bytes:
        """
        The header of the token. It will be ``"<version>.<type>."``.
        For example, ``"v1.local."``.
        """
        return self._header

    @property
    def payload(self) -> t.Union[bytes, dict]:
        """
        The payload of the token which is a decoded binary string. It's not Base64 encoded data.
        """
        return self._payload

    @property
    def footer(self) -> t.Union[bytes, dict]:
        """
        The footer of the token which is a decoded binary string. It's not Base64 encoded data.
        """
        return self._footer

    @property
    def error(self) -> t.Union[Exception, None]:
        """
        If ``is_verified`` is ``True``, this property will be specified.
        """
        return self._error


current_paseto: Token = LocalProxy(lambda: _get_token())  # type: ignore


def paseto_required():
    """ """

    def wrapper(func):
        @wraps(func)
        def decorated_view(*args, **kwargs):

            if not hasattr(current_app, "paseto_verifier"):
                raise ConfigError("paseto_verifier is not set in the current_app.")

            if not current_paseto.is_verified:
                return current_app.paseto_verifier.verification_error_handler_callback()

            if not hasattr(current_app, "ensure_sync"):
                return func(*args, **kwargs)
            # ensure_sync is available in Flask >= 2.0
            return current_app.ensure_sync(func)(*args, **kwargs)

        return decorated_view

    return wrapper


def _get_token() -> Token:
    """ """
    if has_request_context() and not hasattr(_request_ctx_stack.top, "paseto"):
        current_app.paseto_verifier._load_and_verify()  # type: ignore

    return getattr(_request_ctx_stack.top, "paseto", None)
